/*
  I2C Master Library using USI (Fast mode)

  T. Nakagawa
  2010-05-13
*/


#include <avr/io.h>
#include <util/delay.h>
#include "usi_i2c.h"


#define USI_PORT PORTB
#define USI_DDR DDRB
#define SCL_PIN PB7
#define SDA_PIN PB5

#define STAT(bit) (0xf0 | (16 - 2 * bit))


static void transfer(uint8_t stat);
static void delay(void);


/*
  Initialize the interface
*/
void i2c_init(void) {
  USICR |= (_BV(USIWM1) | _BV(USICS1) | _BV(USICLK));

  USI_DDR |= (_BV(SCL_PIN) | _BV(SDA_PIN));
  USIDR = 0xff;
  USI_PORT |= (_BV(SCL_PIN) | _BV(SDA_PIN));
  delay();

  return;
}


/*
  Send/Receive data (recv0̎͑MC1͎̎M)
*/
uint8_t i2c_sendrecv(uint8_t addr, uint8_t size, uint8_t *data, uint8_t recv) {
  /* Start condition */
  USI_PORT &= ~_BV(SDA_PIN);
  delay();
  USI_PORT &= ~_BV(SCL_PIN);
  delay();

  /* Write address and read ACK */
  USI_PORT |= _BV(SDA_PIN);
  USIDR = addr;
  transfer(STAT(8));
  USI_DDR &= ~_BV(SDA_PIN);
  transfer(STAT(1));
  if (USIDR & 0x01) goto stop;	/* NACK */

  /* Send/Receive data */
  do {
    /* Receive data */
    if (recv) {
      USI_DDR &= ~_BV(SDA_PIN);
      transfer(STAT(8));
      *data++ = USIDR;
      USI_DDR |= _BV(SDA_PIN);
      USIDR = (size == 1) ? 0xff : 0x00;
    } else {
      /* Send data */
      USI_DDR |= _BV(SDA_PIN);
      USIDR = *data++;
      transfer(STAT(8));
      USI_DDR &= ~_BV(SDA_PIN);
    }
    transfer(STAT(1));
    if (USIDR & 0x01) goto stop;	/* NACK */
  } while (--size);

  /* Stop condition */
stop:
  USI_DDR |= _BV(SDA_PIN);
  USI_PORT &= ~_BV(SDA_PIN);
  USIDR = 0xff;
  USI_PORT |= _BV(SCL_PIN);
  delay();
  USI_PORT |= _BV(SDA_PIN);
  delay();

  return (size != 0);
}


/*
  Transfer data
*/
static void transfer(uint8_t stat) {
  USISR = stat;
  do {
    USICR |= _BV(USITC);
    delay();
    USICR |= _BV(USITC);
    delay();
  } while (!(USISR & _BV(USIOIF)));

  return;
}


static void delay(void) {
  _delay_us(1.3);

  return;
}
